import re


class Collection:
    """
    Represents a collection of items that can be queried via the API.  This is
    an abstract base class that should be extended
    """

    TIMESTAMP = re.compile(r"[0-9]{4}-?[0-9]{2}-?[0-9]{2}T?" +
                           r"[0-9]{2}:?[0-9]{2}:?[0-9]{2}Z?")
    """
    A regular expression for matching the API's timestamp format.
    """

    DIRECTIONS = {'newer', 'older'}
    """
    A set of potential direction names.
    """

    def __init__(self, session):
        """
        :Parameters:
            session : `mw.api.Session`
                An api session to use for post & get.
        """
        self.session = session
    
    def _check_direction(self, direction):
        if direction is None:
            return direction
        else:
            direction = str(direction)

            assert direction in {None} | self.DIRECTIONS, \
                "Direction must be one of {0}".format(self.DIRECTIONS)

            return direction

    def _check_timestamp(self, timestamp):
        if timestamp is None:
            return timestamp
        else:
            timestamp = str(timestamp)

            if not self.TIMESTAMP.match(timestamp):
                raise TypeError(
                    "{0} is not formatted like ".format(repr(timestamp)) +
                    "a MediaWiki timestamp."
                )

            return timestamp

    def _items(self, items, none=True, levels=None, type=lambda val: val):

        if none and items is None:
            return None
        else:
            items = {str(type(item)) for item in items}

            if levels is not None:
                levels = {str(level) for level in levels}

                assert len(items - levels) == 0, \
                    "items {0} not in levels {1}".format(
                        items - levels, levels)

            return "|".join(items)



from ...types import serializable
from ...util import none_or

from ..errors import MalformedXML
from .revision import Revision
from .redirect import Redirect


class Page(serializable.Type):
    """
    Page meta data and a :class:`~mw.xml_dump.Revision` iterator.  Instances of
    this class can be called as iterators directly.  E.g.

    .. code-block:: python

        page = mw.xml_dump.Page( ... )

        for revision in page:
            print("{0} {1}".format(revision.id, page_id))

    """
    __slots__ = (
        'id',
        'title',
        'namespace',
        'redirect',
        'restrictions',
        'revisions'
    )

    def __init__(self, id, title, namespace, redirect, restrictions, revisions):
        self.id = none_or(id, int)
        """
        Page ID : `int`
        """

        self.title = none_or(title, str)
        """
        Page title (namespace excluded) : `str`
        """

        self.namespace = none_or(namespace, int)
        """
        Namespace ID : `int`
        """

        self.redirect = none_or(redirect, Redirect)
        """
        Page is currently redirect? : :class:`~mw.xml_dump.Redirect` | `None`
        """

        self.restrictions = serializable.List.deserialize(restrictions)
        """
        A list of page editing restrictions (empty unless restrictions are specified) : list( `str` )
        """

        # Should be a lazy generator
        self.__revisions = revisions

    def __iter__(self):
        return self.__revisions

    def __next__(self):
        return next(self.__revisions)

    @classmethod
    def load_revisions(cls, first_revision, element):
        yield Revision.from_element(first_revision)

        for sub_element in element:
            tag = sub_element.tag

            if tag == "revision":
                yield Revision.from_element(sub_element)
            else:
                raise MalformedXML("Expected to see 'revision'.  " +
                                   "Instead saw '{0}'".format(tag))

    @classmethod
    def from_element(cls, element):
        title = None
        namespace = None
        id = None
        redirect = None
        restrictions = []

        first_revision = None

        # Consume each of the elements until we see <id> which should come last.
        for sub_element in element:
            tag = sub_element.tag
            if tag == "title":
                title = sub_element.text
            elif tag == "ns":
                namespace = sub_element.text
            elif tag == "id":
                id = int(sub_element.text)
            elif tag == "redirect":
                redirect = Redirect.from_element(sub_element)
            elif tag == "restrictions":
                restrictions.append(sub_element.text)
            elif tag == "revision":
                first_revision = sub_element
                break
            # Assuming that the first revision seen marks the end of page
            # metadata.  I'm not too keen on this assumption, so I'm leaving
            # this long comment to warn whoever ends up maintaining this.
            else:
                raise MalformedXML("Unexpected tag found when processing " +
                                   "a <page>: '{0}'".format(tag))

        # Assuming that I got here by seeing a <revision> tag.  See verbose
        # comment above.
        revisions = cls.load_revisions(first_revision, element)

        return cls(id, title, namespace, redirect, restrictions, revisions)



import logging

from ..errors import MalformedResponse
from .collection import Collection

logger = logging.getLogger("mw.api.collections.site_info")


class SiteInfo(Collection):
    """
    General information about the site.
    """

    PROPERTIES = {'general', 'namespaces', 'namespacealiases',
                  'specialpagealiases', 'magicwords', 'interwikimap',
                  'dbrepllag', 'statistics', 'usergroups', 'extensions',
                  'fileextensions', 'rightsinfo', 'languages', 'skins',
                  'extensiontags', 'functionhooks', 'showhooks',
                  'variables', 'protocols'}

    FILTERIW = {'local', '!local'}

    def query(self, properties=None, filteriw=None, showalldb=None,
              numberinggroup=None, inlanguagecode=None):
        """
        General information about the site.
        See `<https://www.mediawiki.org/wiki/API:Meta#siteinfo_.2F_si>`_

        :Parameters:
            properties: set(str)
                Which sysinfo properties to get:

                * general               - Overall system information
                * namespaces            - List of registered namespaces and their canonical names
                * namespacealiases      - List of registered namespace aliases
                * specialpagealiases    - List of special page aliases
                * magicwords            - List of magic words and their aliases
                * statistics            - Returns site statistics
                * interwikimap          - Returns interwiki map (optionally filtered, (optionally localised by using siinlanguagecode))
                * dbrepllag             - Returns database server with the highest replication lag
                * usergroups            - Returns user groups and the associated permissions
                * extensions            - Returns extensions installed on the wiki
                * fileextensions        - Returns list of file extensions allowed to be uploaded
                * rightsinfo            - Returns wiki rights (license) information if available
                * restrictions          - Returns information on available restriction (protection) types
                * languages             - Returns a list of languages MediaWiki supports(optionally localised by using siinlanguagecode)
                * skins                 - Returns a list of all enabled skins
                * extensiontags         - Returns a list of parser extension tags
                * functionhooks         - Returns a list of parser function hooks
                * showhooks             - Returns a list of all subscribed hooks (contents of $wgHooks)
                * variables             - Returns a list of variable IDs
                * protocols             - Returns a list of protocols that are allowed in external links.
                * defaultoptions        - Returns the default values for user preferences.
            filteriw : str
                "local" or "!local" Return only local or only nonlocal entries of the interwiki map
            showalldb : bool
                List all database servers, not just the one lagging the most
            numberingroup : bool
                Lists the number of users in user groups
            inlanguagecode : bool
                Language code for localised language names (best effort, use CLDR extension)
  """

        siprop = self._items(properties, levels=self.PROPERTIES)

        doc = self.session.get(
            {
                'action': "query",
                'meta': "siteinfo",
                'siprop': siprop,
                'sifilteriw': filteriw,
                'sishowalldb': showalldb,
                'sinumberinggroup': numberinggroup,
                'siinlanguagecode': inlanguagecode
            }
        )

        try:
            return doc['query']
        except KeyError as e:
            raise MalformedResponse(str(e), doc)



from . import autovivifying


class Circle(list):
    def __init__(self, maxsize, iterable=None):
        self._maxsize = int(maxsize)
        list.__init__(self, [None] * maxsize)
        self._size = 0
        self._pointer = 0

        if iterable is not None:
            self.extend(iterable)

    def state(self):
        return list(list.__iter__(self))

    def _internalize(self, index):
        if self._size < self._maxsize:
            return index
        else:
            return (self._pointer + index) % self._maxsize

    def __iter__(self):
        for i in range(0, self._size):
            yield list.__getitem__(self, self._internalize(i))

    def __reversed__(self):
        for i in range(self._size - 1, -1, -1):
            yield list.__getitem__(self, self._internalize(i))

    def pop(self, index=None):
        raise NotImplementedError()

    def __len__(self):
        return self._size

    def __getitem__(self, index):
        return list.__getitem__(self, self._internalize(index))

    def append(self, value):
        # Get the old value
        old_value = list.__getitem__(self, self._pointer)

        # Update internal list
        list.__setitem__(self, self._pointer, value)

        # Update state
        self._pointer = (self._pointer + 1) % self._maxsize
        self._size = min(self._maxsize, self._size + 1)

        # If we overwrote a value, yield it.
        return old_value

    def extend(self, values):
        for value in values:
            expectorate = self.append(value)
            if expectorate is not None or self._size == self._maxsize:
                yield expectorate


class HistoricalMap(autovivifying.Dict):
    '''
    A datastructure for efficiently storing and retrieving a
    limited number of historical records.

    TODO: Rename this to FIFOCache
    '''

    def __init__(self, *args, maxlen, **kwargs):
        '''Maxlen specifies the maximum amount of history to keep'''
        super().__init__(self, *args, vivifier=lambda k: [], **kwargs)

        self._circle = Circle(maxlen)  # List to preserve order for history

    def __iter__(self):
        return iter(self._circle)

    def __setitem__(self, key, value):
        '''Adds a new key-value pair. Returns any discarded values.'''

        # Add to history circle and catch expectorate
        expectorate = self._circle.append((key, value))

        autovivifying.Dict.__getitem__(self, key).append(value)

        if expectorate is not None:
            old_key, old_value = expectorate
            autovivifying.Dict.__getitem__(self, old_key).pop(0)
            if len(autovivifying.Dict.__getitem__(self, old_key)) == 0:
                autovivifying.Dict.__delitem__(self, old_key)

            return (old_key, old_value)

    def insert(self, key, value):
        return self.__setitem__(key, value)

    def __getitem__(self, key):
        if key in self:
            return autovivifying.Dict.__getitem__(self, key)[-1]
        else:
            raise KeyError(key)

    def get(self, key):
        '''Gets the most recently added value for a key'''
        return self.__getitem__(key)

    def up_to(self, key):
        '''Gets the recently inserted values up to a key'''
        for okey, ovalue in reversed(self._circle):
            if okey == key:
                break
            else:
                yield ovalue

    def last(self):
        return self.circle[-1]



import logging
import time

from ...types import Timestamp
from ...util import none_or
from .collection import Collection

logger = logging.getLogger("mw.database.collections.pages")


class RecentChanges(Collection):
    # (https://www.mediawiki.org/wiki/Manual:Recentchanges_table)
    TYPES = {
        'edit': 0,  # edit of existing page
        'new': 1,  # new page
        'move': 2,  # Marked as obsolete
        'log': 3,  # log action (introduced in MediaWiki 1.2)
        'move_over_redirect': 4,  # Marked as obsolete
        'external': 5  # An external recent change. Primarily used by Wikidata
    }

    def listen(self, last=None, types=None, max_wait=5):
        """
        Listens to the recent changes table.  Given no parameters, this function
        will return an iterator over the entire recentchanges table and then
        continue to "listen" for new changes to come in every 5 seconds.

        :Parameters:
            last : dict
                a recentchanges row to pick up after
            types : set ( str )
                a set of recentchanges types to filter for
            max_wait : float
                the maximum number of seconds to wait between repeated queries

        :Returns:
            A never-ending iterator over change rows.
        """
        while True:
            if last is not None:
                after = last['rc_timestamp']
                after_id = last['rc_id']
            else:
                after = None
                after_id = None

            start = time.time()
            rcs = self.query(after=after, after_id=after_id, direction="newer")

            count = 0
            for rc in rcs:
                yield rc
                count += 1

            time.sleep(max_wait - (time.time() - start))

    def query(self, before=None, after=None, before_id=None, after_id=None,
              types=None, direction=None, limit=None):
        """
        Queries the ``recentchanges`` table.  See
        `<https://www.mediawiki.org/wiki/Manual:Recentchanges_table>`_

        :Parameters:
            before : :class:`mw.Timestamp`
                The maximum timestamp
            after : :class:`mw.Timestamp`
                The minimum timestamp
            before_id : int
                The minimum ``rc_id``
            after_id : int
                The maximum ``rc_id``
            types : set ( str )
                Which types of changes to return?

                * ``edit`` -- Edits to existing pages
                * ``new`` -- Edits that create new pages
                * ``move`` -- (obsolete)
                * ``log`` -- Log actions (introduced in MediaWiki 1.2)
                * ``move_over_redirect`` -- (obsolete)
                * ``external`` -- An external recent change. Primarily used by Wikidata

            direction : str
                "older" or "newer"
            limit : int
                limit the number of records returned
        """
        before = none_or(before, Timestamp)
        after = none_or(after, Timestamp)
        before_id = none_or(before_id, int)
        after_id = none_or(after_id, int)
        types = none_or(types, levels=self.TYPES)
        direction = none_or(direction, levels=self.DIRECTIONS)
        limit = none_or(limit, int)

        query = """
            SELECT * FROM recentchanges
            WHERE 1
        """
        values = []

        if before is not None:
            query += " AND rc_timestamp < %s "
            values.append(before.short_format())
        if after is not None:
            query += " AND rc_timestamp < %s "
            values.append(after.short_format())
        if before_id is not None:
            query += " AND rc_id < %s "
            values.append(before_id)
        if after_id is not None:
            query += " AND rc_id < %s "
            values.append(after_id)
        if types is not None:
            query += " AND rc_type IN ({0}) ".format(
                ",".join(self.TYPES[t] for t in types)
            )

        if direction is not None:
            direction = ("ASC " if direction == "newer" else "DESC ")
            query += " ORDER BY rc_timestamp {0}, rc_id {0}".format(dir)

        if limit is not None:
            query += " LIMIT %s "
            values.append(limit)

        cursor.execute(query, values)
        for row in cursor:
            yield row



from nose.tools import eq_
from ..peekable import Peekable


def test_peekable():
    iterable = range(0, 100)
    iterable = Peekable(iterable)
    expected = list(range(0, 100))

    result = []

    assert not iterable.empty()
    eq_(iterable.peek(), expected[0])
    result.append(next(iterable))

    eq_(iterable.peek(), expected[1])
    result.append(next(iterable))

    result.extend(list(iterable))
    eq_(result, expected)



import io

from ...types import serializable
from ...util import none_or
from ..element_iterator import ElementIterator
from ..errors import MalformedXML
from .namespace import Namespace
from .page import Page


class ConcatinatingTextReader(io.TextIOBase):
    
    def __init__(self, *items):
        self.items = [io.StringIO(i) if isinstance(i, str) else i
                      for i in items]
    
    def read(self, size=-1):
        return "".join(self._read(size))
    
    def readline(self):
        
        if len(self.items) > 0:
            line = self.items[0].readline()
            if line == "": self.items.pop(0)
        else:
            line = ""
        
        return line
    
    def _read(self, size):
        if size > 0:
            while len(self.items) > 0:
                byte_vals = self.items[0].read(size)
                yield byte_vals
                if len(byte_vals) < size:
                    size = size - len(byte_vals) # Decrement bytes
                    self.items.pop(0)
                else:
                    break
        
        else:
            for item in self.items:
                yield item.read()
            
        
    

def concat(*stream_items):
    return ConcatinatingTextReader(*stream_items)


class Iterator(serializable.Type):
    """
    XML Dump Iterator. Dump file meta data and a
    :class:`~mw.xml_dump.Page` iterator.  Instances of this class can be
    called as an iterator directly.  E.g.::

        from mw.xml_dump import Iterator

        # Construct dump file iterator
        dump = Iterator.from_file(open("example/dump.xml"))

        # Iterate through pages
        for page in dump:

            # Iterate through a page's revisions
            for revision in page:

                print(revision.id)

    """
    __slots__ = ('site_name', 'base', 'generator', 'case', 'namespaces',
                 '__pages')

    def __init__(self, site_name=None, base=None, generator=None, case=None,
                 namespaces=None, pages=None):

        self.site_name = none_or(site_name, str)
        """
        The name of the site. : str | `None` (if not specified in the XML)
        """

        self.base = none_or(base, str)
        """
        TODO: ??? : str | `None` (if not specified in the XML)
        """

        self.generator = none_or(generator, str)
        """
        TODO: ??? : str | `None` (if not specified in the XML)
        """

        self.case = none_or(case, str)
        """
        TODO: ??? : str | `None` (if not specified in the XML)
        """

        self.namespaces = none_or(namespaces, list)
        """
        A list of :class:`mw.Namespace` | `None` (if not specified in the XML)
        """

        # Should be a lazy generator of page info
        self.__pages = pages

    def __iter__(self):
        return self.__pages

    def __next__(self):
        return next(self.__pages)

    @classmethod
    def load_namespaces(cls, element):
        namespaces = []
        for sub_element in element:
            tag = sub_element.tag

            if tag == "namespace":
                namespace = Namespace.from_element(sub_element)
                namespaces.append(namespace)
            else:
                assert False, "This should never happen"

        return namespaces

    @classmethod
    def load_site_info(cls, element):

        site_name = None
        base = None
        generator = None
        case = None
        namespaces = {}

        for sub_element in element:
            if sub_element.tag == 'sitename':
                site_name = sub_element.text
            elif sub_element.tag == 'base':
                base = sub_element.text
            elif sub_element.tag == 'generator':
                generator = sub_element.text
            elif sub_element.tag == 'case':
                case = sub_element.text
            elif sub_element.tag == 'namespaces':
                namespaces = cls.load_namespaces(sub_element)

        return site_name, base, generator, case, namespaces

    @classmethod
    def load_pages(cls, element):

        for sub_element in element:
            tag = sub_element.tag

            if tag == "page":
                yield Page.from_element(sub_element)
            else:
                assert MalformedXML("Expected to see 'page'.  " +
                                    "Instead saw '{0}'".format(tag))

    @classmethod
    def from_element(cls, element):

        site_name = None
        base = None
        generator = None
        case = None
        namespaces = None

        # Consume <siteinfo>
        for sub_element in element:
            tag = sub_element.tag
            if tag == "siteinfo":
                site_name, base, generator, case, namespaces = \
                    cls.load_site_info(sub_element)
                break

        # Consume all <page>
        pages = cls.load_pages(element)

        return cls(site_name, base, generator, case, namespaces, pages)

    @classmethod
    def from_file(cls, f):
        element = ElementIterator.from_file(f)
        assert element.tag == "mediawiki"
        return cls.from_element(element)
    
    @classmethod
    def from_page_xml(cls, page_xml):
        header = """
        <mediawiki xmlns="http://www.mediawiki.org/xml/export-0.5/"
                   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                   xsi:schemaLocation="http://www.mediawiki.org/xml/export-0.5/
                     http://www.mediawiki.org/xml/export-0.5.xsd" version="0.5"
                   xml:lang="en">
        <siteinfo>
            <namespaces>
            </namespaces>
        </siteinfo>
        """
        
        footer = "</mediawiki>"
        
        return cls.from_file(concat(header, page_xml, footer))



import heapq


class Heap(list):
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        heapq.heapify(self)

    def pop(self):
        return heapq.heappop(self)

    def push(self, item):
        heapq.heappush(self, item)

    def peek(self):
        return self[0]

    def pushpop(self, item):
        return heapq.heappushpop(self, item)

    def poppush(self, itemp):
        return heapq.replace(self, item)



from nose.tools import eq_

from .. import difference


def test_sequence_matcher():
    t1 = "foobar derp hepl derpl"
    t2 = "fooasldal 3 hepl asl a derpl"

    ops = difference.sequence_matcher(t1, t2)

    eq_("".join(difference.apply(ops, t1, t2)), t2)



import logging
from collections import namedtuple

from ...util import Heap
from ...types import Timestamp
from . import defaults
from .event import Event, unpack_events


logger = logging.getLogger("mw.lib.sessions.cache")

Session = namedtuple("Session", ["user", "events"])
"""
Represents a user session (a cluster over events for a user).  This class
behaves like :class:`collections.namedtuple`.  Note that the datatypes of
`events`, is not specified since those types will depend on the revision data
provided during revert detection.

:Members:
    **user**
        A hashable user identifier : `hashable`
    **events**
        A list of event data : list( `mixed` )
"""


class Cache:
    """
    A cache of recent user session.  Since sessions expire once activities stop
    for at least `cutoff` seconds, this class manages a cache of *active*
    sessions.

    :Parameters:
        cutoff : int
            Maximum amount of time in seconds between session events

    :Example:
        >>> from mw.lib import sessions
        >>>
        >>> cache = sessions.Cache(cutoff=3600)
        >>>
        >>> list(cache.process("Willy on wheels", 100000, {'rev_id': 1}))
        []
        >>> list(cache.process("Walter", 100001, {'rev_id': 2}))
        []
        >>> list(cache.process("Willy on wheels", 100001, {'rev_id': 3}))
        []
        >>> list(cache.process("Walter", 100035, {'rev_id': 4}))
        []
        >>> list(cache.process("Willy on wheels", 103602, {'rev_id': 5}))
        [Session(user='Willy on wheels', events=[{'rev_id': 1}, {'rev_id': 3}])]
        >>> list(cache.get_active_sessions())
        [Session(user='Walter', events=[{'rev_id': 2}, {'rev_id': 4}]), Session(user='Willy on wheels', events=[{'rev_id': 5}])]


    """

    def __init__(self, cutoff=defaults.CUTOFF):
        self.cutoff = int(cutoff)

        self.recently_active = Heap()
        self.active_users = {}

    def process(self, user, timestamp, data=None):
        """
        Processes a user event.

        :Parameters:
            user : `hashable`
                A hashable value to identify a user (`int` or `str` are OK)
            timestamp : :class:`mw.Timestamp`
                The timestamp of the event
            data : `mixed`
                Event meta data

        :Returns:
            A generator of :class:`~mw.lib.sessions.Session` expired after
            processing the user event.
        """
        event = Event(user, Timestamp(timestamp), data)

        for user, events in self._clear_expired(event.timestamp):
            yield Session(user, unpack_events(events))

        # Apply revision
        if event.user in self.active_users:
            events = self.active_users[event.user]
        else:
            events = []
            self.active_users[event.user] = events
            self.recently_active.push((event.timestamp, events))

        events.append(event)

    def get_active_sessions(self):
        """
        Retrieves the active, unexpired sessions.

        :Returns:
            A generator of :class:`~mw.lib.sessions.Session`

        """
        for last_timestamp, events in self.recently_active:
            yield Session(events[-1].user, unpack_events(events))

    def _clear_expired(self, timestamp):

        # Cull old sessions
        while (len(self.recently_active) > 0 and
               timestamp - self.recently_active.peek()[0] >= self.cutoff):

            _, events = self.recently_active.pop()

            if timestamp - events[-1].timestamp >= self.cutoff:
                del self.active_users[events[-1].user]
                yield events[-1].user, events
            else:
                self.recently_active.push((events[-1].timestamp, events))

    def __repr__(self):
        return "%s(%s)".format(self.__class__.__name__, repr(self.cutoff))



from nose.tools import eq_

from ..namespace import Namespace


def test_namespace():
    namespace = Namespace(10, "Foo", canonical="Bar", aliases={'WT'}, case="foobar")

    eq_(namespace.id, 10)
    eq_(namespace.name, "Foo")
    eq_(namespace.canonical, "Bar")
    eq_(namespace.aliases, {'WT'})
    eq_(namespace.case, "foobar")



from .peekable import Peekable


def sequence(*iterables, by=None, compare=None):
    if compare is not None:
        compare = compare
    elif by is not None:
        compare = lambda i1, i2: by(i1) <= by(i2)
    else:
        compare = lambda i1, i2: i1 <= i2

    iterables = [Peekable(it) for it in iterables]

    done = False
    while not done:

        next_i = None

        for i, it in enumerate(iterables):
            if not it.empty():
                if next_i is None or compare(it.peek(), iterables[next_i].peek()):
                    next_i = i

        if next_i is None:
            done = True
        else:
            yield next(iterables[next_i])



from itertools import chain

from . import defaults
from ...types import Timestamp
from ...util import none_or
from .functions import detect


"""
Simple constant used in order to not do weird things with a dummy revision.
"""


def check_row(db, rev_row, **kwargs):
    """
    Checks whether a revision (database row) was reverted (identity) and returns
    a named tuple of Revert(reverting, reverteds, reverted_to).

    :Parameters:
        db : :class:`mw.database.DB`
            A database connection to make use of.
        rev_row : dict
            a revision row containing 'rev_id' and 'rev_page' or 'page_id'
        radius : int
            the maximum number of revisions that can be reverted
        check_archive : bool
            should the archive table be check for reverting revisions?
        before : `Timestamp`
            if set, limits the search for *reverting* revisions to those which were saved before this timestamp
    """

    # extract rev_id, sha1, page_id
    if 'rev_id' in rev_row:
        rev_id = rev_row['rev_id']
    else:
        raise TypeError("rev_row must have 'rev_id'")
    if 'page_id' in rev_row:
        page_id = rev_row['page_id']
    elif 'rev_page' in rev_row:
        page_id = rev_row['rev_page']
    else:
        raise TypeError("rev_row must have 'page_id' or 'rev_page'")

    # run the regular check
    return check(db, rev_id, page_id=page_id, **kwargs)


def check(db, rev_id, page_id=None, radius=defaults.RADIUS, check_archive=False, before=None):

    """
    Checks whether a revision was reverted (identity) and returns a named tuple
    of Revert(reverting, reverteds, reverted_to).

    :Parameters:
        db : `mw.database.DB`
            A database connection to make use of.
        rev_id : int
            the ID of the revision to check
        page_id : int
            the ID of the page the revision occupies (slower if not provided)
        radius : int
            the maximum number of revisions that can be reverted
        check_archive : bool
            should the archive table be check for reverting revisions?
        before : `Timestamp`
            if set, limits the search for *reverting* revisions to those which were saved before this timestamp
    """

    if not hasattr(db, "revisions") and hasattr(db, "all_revisions"):
        raise TypeError("db wrong type.  Expected a mw.database.DB.")

    rev_id = int(rev_id)
    radius = int(radius)
    page_id = none_or(page_id, int)
    check_archive = bool(check_archive)
    before = none_or(before, Timestamp)

    # If we are searching the archive, we'll need to use `all_revisions`.
    if check_archive:
        dbrevs = db.all_revisions
    else:
        dbrevs = db.revisions

    # If we don't have the sha1 or page_id, we're going to need to look them up
    if page_id is None:
        row = dbrevs.get(rev_id=rev_id)
        page_id = row['rev_page']

    # Load history and current rev
    current_and_past_revs = list(dbrevs.query(
        page_id=page_id,
        limit=radius + 1,
        before_id=rev_id + 1,  # Ensures that we capture the current revision
        direction="older"
    ))

    try:
        # Extract current rev and reorder history
        current_rev, past_revs = (
            current_and_past_revs[0],  # Current rev is the first one returned
            reversed(current_and_past_revs[1:])  # The rest are past revs, but they are in the wrong order
        )
    except IndexError:
        # Only way to get here is if there isn't enough history.  Couldn't be
        # reverted.  Just return None.
        return None

    # Load future revisions
    future_revs = dbrevs.query(
        page_id=page_id,
        limit=radius,
        after_id=rev_id,
        before=before,
        direction="newer"
    )

    # Convert to an iterable of (checksum, rev) pairs for detect() to consume
    checksum_revisions = chain(
        ((rev['rev_sha1'], rev) for rev in past_revs),
        [(current_rev['rev_sha1'], current_rev)],
        ((rev['rev_sha1'], rev) for rev in future_revs)
    )

    for revert in detect(checksum_revisions, radius=radius):
        # Check that this is a relevant revert
        if rev_id in [rev['rev_id'] for rev in revert.reverteds]:
            return revert

    return None



class FileTypeError(Exception):
    """
    Thrown when an XML dump file is not of an expected type.
    """
    pass


class MalformedXML(Exception):
    """
    Thrown when an XML dump file is not formatted as expected.
    """
    pass



from .. import reverts
from ...util import none_or
from .state import State


def track(session, rev_id, page_id=None, revert_radius=reverts.defaults.RADIUS,
          future_revisions=reverts.defaults.RADIUS, properties=None):
    """
    Computes a persistence score for a revision by processing the revisions
    that took place around it.

    :Parameters:
        session : :class:`mw.api.Session`
            An API session to make use of
        rev_id : int
            the ID of the revision to check
        page_id : int
            the ID of the page the revision occupies (slower if not provided)
        revert_radius : int
            the maximum number of revisions that can be reverted
    """

    if not hasattr(session, "revisions"):
        raise TypeError("session is wrong type.  Expected a mw.api.Session.")

    rev_id = int(rev_id)
    page_id = none_or(page_id, int)
    revert_radius = int(revert_radius)
    properties = set(properties) if properties is not None else set()
    

    # If we don't have the page_id, we're going to need to look them up
    if page_id is None:
        rev = session.revisions.get(rev_id, properties={'ids'})
        page_id = rev['page']['pageid']
    
    # Load history and current rev
    current_and_past_revs = list(session.revisions.query(
        pageids={page_id},
        limit=revert_radius + 1,
        start_id=rev_id + 1,  # Ensures that we capture the current revision
        direction="older",
        properties={'ids', 'timestamp', 'content', 'sha1'} | properties
    ))

    try:
        # Extract current rev and reorder history
        current_rev, past_revs = (
            current_and_past_revs[0],  # Current rev is the first one returned
            reversed(current_and_past_revs[1:])  # The rest are past revs, but they are in the wrong order
        )
    except IndexError:
        # Only way to get here is if there isn't enough history.  Couldn't be
        # reverted.  Just return None.
        return None

    # Load future revisions
    future_revs = session.revisions.query(
        pageids={page_id},
        limit=future_revisions,
        start_id=rev_id+1,
        direction="newer",
        properties={'ids', 'timestamp', 'content', 'sha1'} | properties
    )
    
    state = State(revert_radius=revert_radius)
    
    # Process old revisions
    for rev in past_revs:
        state.process(rev.get('*', ""), rev, rev.get('sha1'))
    
    # Process current revision
    _, tokens_added, _ = state.process(current_rev.get('*'), current_rev,
                                         current_rev.get('sha1'))
    
    # Process new revisions
    future_revs = list(future_revs)
    for rev in future_revs:
        state.process(rev.get('*', ""), rev, rev.get('sha1'))
        future_rev_count += 1
    
    
    return current_rev, tokens_added, future_revs

score = track



import io

from nose.tools import eq_, raises
from ..map import map


SAMPLE_XML = """
<mediawiki xmlns="http://www.mediawiki.org/xml/export-0.8/"
           xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
           xsi:schemaLocation="http://www.mediawiki.org/xml/export-0.8/
           http://www.mediawiki.org/xml/export-0.8.xsd"
           version="0.8" xml:lang="en">
  <siteinfo>
    <sitename>Wikipedia</sitename>
    <base>http://en.wikipedia.org/wiki/Main_Page</base>
    <generator>MediaWiki 1.22wmf2</generator>
    <case>first-letter</case>
    <namespaces>
      <namespace key="0" case="first-letter" />
      <namespace key="1" case="first-letter">Talk</namespace>
    </namespaces>
  </siteinfo>
  <page>
    <title>Foo</title>
    <ns>0</ns>
    <id>1</id>
    <revision>
      <id>1</id>
      <timestamp>2004-08-09T09:04:08Z</timestamp>
      <contributor>
        <username>Gen0cide</username>
        <id>92182</id>
      </contributor>
      <text xml:space="preserve">Revision 1 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
    <revision>
      <id>2</id>
      <timestamp>2004-08-10T09:04:08Z</timestamp>
      <contributor>
        <ip>222.152.210.109</ip>
      </contributor>
      <text xml:space="preserve">Revision 2 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <comment>Comment 2</comment>
      <format>text/x-wiki</format>
    </revision>
  </page>
  <page>
    <title>Bar</title>
    <ns>1</ns>
    <id>2</id>
    <revision>
      <id>3</id>
      <timestamp>2004-08-11T09:04:08Z</timestamp>
      <contributor>
        <ip>222.152.210.22</ip>
      </contributor>
      <text xml:space="preserve">Revision 3 text</text>
      <sha1>g9chqqg94myzq11c56ixvq7o1yg75n9</sha1>
      <model>wikitext</model>
      <format>text/x-wiki</format>
    </revision>
  </page>
</mediawiki>"""


def test_map():
    f = io.StringIO(SAMPLE_XML)

    def process_dump(dump, path):
        for page in dump:
            count = 0
            for rev in page:
                count += 1

            yield {'page_id': page.id, 'revisions': count}

    pages = 0
    for doc in map([f], process_dump):
        page_id = doc['page_id']
        revisions = doc['revisions']
        if page_id == 1:
            eq_(revisions, 2)
        elif page_id == 2:
            eq_(revisions, 1)
        else:
            assert False

        pages += 1

    eq_(pages, 2)


def test_dict_yield():
    def test_map():
        f = io.StringIO(SAMPLE_XML)

        def process_dump(dump, path):
            for page in dump:
                count = 0
                for rev in page:
                    count += 1

                yield {'page_id': page.id, 'revisions': count}

        pages = 0
        for doc in map([f], process_dump):
            page_id = doc['page_id']
            revisions = doc['revisions']
            if page_id == 1:
                eq_(revisions, 2)
            elif page_id == 2:
                eq_(revisions, 1)
            else:
                assert False

            pages += 1

        eq_(pages, 2)


@raises(TypeError)
def test_map_error():
    f = io.StringIO(SAMPLE_XML)

    def process_dump(dump, path):
        for page in dump:

            if page.id == 2:
                raise TypeError("Fake error")

    pages = 0
    for doc in map([f], process_dump):
        page_id = doc['page_id']


def test_map_error_handler():
    f = io.StringIO(SAMPLE_XML)

    def process_dump(dump, path, handle_error=lambda exp, stack: None):
        for page in dump:
            count = 0

            for rev in page:
                count += 1

            if count > 2:
                raise TypeError("Fake type error.")

            yield {'page_id': page.id, 'revisions': count}

    pages = 0
    for doc in map([f], process_dump):
        page_id = doc['page_id']
        revisions = doc['revisions']
        if page_id == 1:
            eq_(revisions, 2)
        elif page_id == 2:
            eq_(revisions, 1)
        else:
            assert False

        pages += 1

    eq_(pages, 2)



import logging
import time

from ...types import Timestamp
from ...util import none_or
from .collection import Collection

logger = logging.getLogger("mw.database.collections.users")


class Users(Collection):
    CREATION_ACTIONS = {'newusers', 'create', 'create2', 'autocreate',
                        'byemail'}

    def get(self, user_id=None, user_name=None):
        """
        Gets a single user row from the database.  Raises a :class:`KeyError`
        if a user cannot be found.

        :Parameters:
            user_id : int
                User ID
            user_name : str
                User's name

        :Returns:
            A user row.
        """
        user_id = none_or(user_id, int)
        user_name = none_or(user_name, str)

        query = """
            SELECT user.*
            FROM user
        """
        values = []

        if user_id is not None:
            query += """
                WHERE user_id = %s
            """
            values.append(user_id)

        elif user_name is not None:
            query += """
                WHERE user_name = %s
            """
            values.append(user_name)

        else:
            raise TypeError("Must specify a user identifier.")

        cursor = self.db.shared_connection.cursor()
        cursor.execute(
            query,
            values
        )

        for row in cursor:
            return row

        raise KeyError(user_id if user_id is not None else user_name)

    def query(self, registered_before=None, registered_after=None,
              before_id=None, after_id=None, limit=None,
              direction=None, self_created_only=False):
        """
        Queries users based on various filtering parameters.

        :Parameters:
            registered_before : :class:`mw.Timestamp`
                A timestamp to search before (inclusive)
            registered_after : :class:`mw.Timestamp`
                A timestamp to search after (inclusive)
            before_id : int
                A user_id to search before (inclusive)
            after_id : int
                A user_ud to search after (inclusive)
            direction : str
                "newer" or "older"
            limit : int
                Limit the results to at most this number
            self_creations_only : bool
                limit results to self_created user accounts

        :Returns:
            an iterator over ``user`` table rows
        """
        start_time = time.time()

        registered_before = none_or(registered_before, Timestamp)
        registered_after = none_or(registered_after, Timestamp)
        before_id = none_or(before_id, str)
        after_id = none_or(after_id, str)
        direction = none_or(direction, levels=self.DIRECTIONS)
        limit = none_or(limit, int)
        self_created_only = bool(self_created_only)

        query = """
            SELECT user.*
            FROM user
        """
        values = []

        if self_created_only:
            query += """
                INNER JOIN logging ON
                    log_user = user_id
                    log_type = "newusers" AND
                    log_action = "create"
            """

        query += "WHERE 1 "

        if registered_before is not None:
            query += "AND user_registration <= %s "
            values.append(registered_before.short_format())
        if registered_after is not None:
            query += "AND user_registration >= %s "
            values.append(registered_after.short_format())
        if before_id is not None:
            query += "AND user_id <= %s "
            values.append(before_id)
        if after_id is not None:
            query += "AND user_id >= %s "
            values.append(after_id)

        query += "GROUP BY user_id "  # In case of duplicate log events
        
        if direction is not None:
            if registered_before is not None or registered_after is not None:
                if direction == "newer":
                    query += "ORDER BY user_registration ASC "
                else:
                    query += "ORDER BY user_registration DESC "
            else:
                if direction == "newer":
                    query += "ORDER BY user_id ASC "
                else:
                    query += "ORDER BY user_id DESC "

        if limit is not None:
            query += "LIMIT %s "
            values.append(limit)

        cursor = self.db.shared_connection.cursor()
        cursor.execute(query, values)

        count = 0
        for row in cursor:
            yield row
            count += 1

        logger.debug("%s users queried in %s seconds" % (count, time.time() - start_time))



from difflib import SequenceMatcher


def sequence_matcher(old, new):
    """
    Generates a sequence of operations using :class:`difflib.SequenceMatcher`.

    :Parameters:
        old : list( `hashable` )
            Old tokens
        new : list( `hashable` )
            New tokens

    Returns:
        Minimal operations needed to convert `old` to `new`
    """
    sm = SequenceMatcher(None, list(old), list(new))
    return sm.get_opcodes()


def apply(ops, old, new):
    """
    Applies operations (delta) to copy items from `old` to `new`.

    :Parameters:
        ops : list((op, a1, a2, b1, b2))
            Operations to perform
        old : list( `hashable` )
            Old tokens
        new : list( `hashable` )
            New tokens
    :Returns:
        An iterator over elements matching `new` but copied from `old`
    """
    for code, a_start, a_end, b_start, b_end in ops:
        if code == "insert":
            for t in new[b_start:b_end]:
                yield t
        elif code == "replace":
            for t in new[b_start:b_end]:
                yield t
        elif code == "equal":
            for t in old[a_start:a_end]:
                yield t
        elif code == "delete":
            pass
        else:
            assert False, \
                "encounted an unrecognized operation code: " + repr(code)



from nose.tools import eq_

from ..functions import normalize


def test_normalize():
    eq_("Foobar", normalize("Foobar"))  # Same
    eq_("Foobar", normalize("foobar"))  # Capitalize
    eq_("FooBar", normalize("fooBar"))  # Late capital
    eq_("Foo_bar", normalize("Foo bar"))  # Space



from itertools import chain


class Type:
    def __eq__(self, other):
        if other is None:
            return False
        try:
            for key in self.keys():
                if getattr(self, key) != getattr(other, key):
                    return False

            return True
        except KeyError:
            return False

    def __neq__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(
                "%s=%r" % (k, v) for k, v in self.items()
            )
        )

    def items(self):
        for key in self.keys():
            yield key, getattr(self, key)

    def keys(self):
        return (
            key for key in
            chain(getattr(self, "__slots__", []), self.__dict__.keys())
            if key[:2] != "__"
        )

    def serialize(self):
        return dict(
            (k, self._serialize(v))
            for k, v in self.items()
        )

    def _serialize(self, value):
        if hasattr(value, "serialize"):
            return value.serialize()
        else:
            return value

    @classmethod
    def deserialize(cls, doc_or_instance):
        if isinstance(doc_or_instance, cls):
            return doc_or_instance
        else:
            return cls(**doc_or_instance)


class Dict(dict, Type):
    def serialize(self):
        return {k: self._serialize(v) for k, v in self.items()}

    @staticmethod
    def deserialize(d, value_deserializer=lambda v: v):
        if isinstance(d, Dict):
            return d
        else:
            return Dict((k, value_deserializer(v)) for k, v in d.items())


class Set(set, Type):
    def serialize(self):
        return [self._serialize(v) for v in self]

    @staticmethod
    def deserialize(s, value_deserializer=lambda v: v):

        if isinstance(s, Set):
            return s
        else:
            return Set(value_deserializer(v) for v in s)


class List(list, Type):
    def serialize(self):
        return list(self._serialize(v) for v in self)

    @staticmethod
    def deserialize(l, value_deserializer=lambda v: v):

        if isinstance(l, List):
            return l
        else:
            return List(value_deserializer(v) for v in l)



from .peekable import Peekable


def group(it, by=lambda i: i):
    return aggregate(it, by)


def aggregate(it, by=lambda i: i):
    it = Peekable(it)

    def chunk(it, by):
        identifier = by(it.peek())
        while not it.empty():
            if identifier == by(it.peek()):
                yield next(it)
            else:
                break

    while not it.empty():
        yield (by(it.peek()), chunk(it, by))



class DocError(Exception):
    def __init__(self, message, doc):
        super().__init__(message)

        self.doc = doc
        """
        The document returned by the API that brought about this error.
        """


class APIError(DocError):
    def __init__(self, doc):
        
        code = doc.get('error', {}).get('code')
        message = doc.get('error', {}).get('message')

        super().__init__("{0}:{1}".format(code, message), doc)

        self.code = code
        """
        The error code returned by the api -- if available.
        """

        self.message = message
        """
        The error message returned by the api -- if available.
        """

class AuthenticationError(APIError):
    def __init__(self, doc):
        result = doc['login']['result']
        super().__init__(doc)

        self.result = result
        """
        The result code of an authentication attempt.
        """


class MalformedResponse(DocError):
    def __init__(self, key, doc):

        super().__init__("Expected to find '{0}' in result.".format(key), doc)

        self.key = key
        """
        The expected, but missing key from the API call.
        """



import logging
import re

from ...util import none_or
from ..errors import MalformedResponse
from .collection import Collection

logger = logging.getLogger("mw.api.collections.recent_changes")


class RecentChanges(Collection):
    """
    Recent changes (revisions, page creations, registrations, moves, etc.)
    """

    RCCONTINUE = re.compile(r"([0-9]{4}-[0-9]{2}-[0-9]{2}T" +
                            r"[0-9]{2}:[0-9]{2}:[0-9]{2}Z|" +
                            r"[0-9]{14})" +
                            r"\|[0-9]+")

    PROPERTIES = {'user', 'userid', 'comment', 'timestamp', 'title',
                  'ids', 'sizes', 'redirect', 'flags', 'loginfo',
                  'tags', 'sha1'}

    SHOW = {'minor', '!minor', 'bot', '!bot', 'anon', '!anon',
            'redirect', '!redirect', 'patrolled', '!patrolled'}
    
    TYPES = {'edit', 'external', 'new', 'log'}
    
    DIRECTIONS = {'newer', 'older'}

    MAX_CHANGES = 50

    def _check_rccontinue(self, rccontinue):
        if rccontinue is None:
            return None
        elif self.RCCONTINUE.match(rccontinue):
            return rccontinue
        else:
            raise TypeError(
                "rccontinue {0} is not formatted correctly ".format(rccontinue) +
                "'%Y-%m-%dT%H:%M:%SZ|<last_rcid>'"
            )

    def query(self, *args, limit=None, **kwargs):
        """
        Enumerate recent changes.
        See `<https://www.mediawiki.org/wiki/API:Recentchanges>`_

        :Parameters:
            start : :class:`mw.Timestamp`
                The timestamp to start enumerating from
            end : :class:`mw.Timestamp`
                The timestamp to end enumerating
            direction :
                "newer" or "older"
            namespace : int
                Filter log entries to only this namespace(s)
            user : str
                Only list changes by this user
            excludeuser : str
                Don't list changes by this user
            tag : str
                Only list changes tagged with this tag
            properties : set(str)
                Include additional pieces of information

                * user           - Adds the user responsible for the edit and tags if they are an IP
                * userid         - Adds the user id responsible for the edit
                * comment        - Adds the comment for the edit
                * parsedcomment  - Adds the parsed comment for the edit
                * flags          - Adds flags for the edit
                * timestamp      - Adds timestamp of the edit
                * title          - Adds the page title of the edit
                * ids            - Adds the page ID, recent changes ID and the new and old revision ID
                * sizes          - Adds the new and old page length in bytes
                * redirect       - Tags edit if page is a redirect
                * patrolled      - Tags patrollable edits as being patrolled or unpatrolled
                * loginfo        - Adds log information (logid, logtype, etc) to log entries
                * tags           - Lists tags for the entry
                * sha1           - Adds the content checksum for entries associated with a revision

            token : set(str)
                Which tokens to obtain for each change

                * patrol

            show : set(str)
                Show only items that meet this criteria. For example, to see
                only minor edits done by logged-in users, set
                show={'minor', '!anon'}.

                * minor
                * !minor
                * bot
                * !bot
                * anon
                * !anon
                * redirect
                * !redirect
                * patrolled
                * !patrolled
                * unpatrolled
            limit : int
                How many total changes to return
            type : set(str)
                Which types of changes to show

                * edit
                * external
                * new
                * log

            toponly : bool
                Only list changes which are the latest revision
            rccontinue : str
                Use this to continue loading results from where you last left off
        """
        limit = none_or(limit, int)

        changes_yielded = 0
        done = False
        while not done:

            if limit is None:
                kwargs['limit'] = self.MAX_CHANGES
            else:
                kwargs['limit'] = min(limit - changes_yielded, self.MAX_CHANGES)

            rc_docs, rccontinue = self._query(*args, **kwargs)

            for doc in rc_docs:
                yield doc
                changes_yielded += 1

                if limit is not None and changes_yielded >= limit:
                    done = True
                    break

            if rccontinue is not None and len(rc_docs) > 0:

                kwargs['rccontinue'] = rccontinue
            else:
                done = True

    def _query(self, start=None, end=None, direction=None, namespace=None,
               user=None, excludeuser=None, tag=None, properties=None,
               token=None, show=None, limit=None, type=None,
               toponly=None, rccontinue=None):

        params = {
            'action': "query",
            'list': "recentchanges"
        }

        params['rcstart'] = none_or(start, str)
        params['rcend'] = none_or(end, str)

        assert direction in {None} | self.DIRECTIONS, \
            "Direction must be one of {0}".format(self.DIRECTIONS)

        params['rcdir'] = direction
        params['rcnamespace'] = none_or(namespace, int)
        params['rcuser'] = none_or(user, str)
        params['rcexcludeuser'] = none_or(excludeuser, str)
        params['rctag'] = none_or(tag, str)
        params['rcprop'] = self._items(properties, levels=self.PROPERTIES)
        params['rctoken'] = none_or(tag, str)
        params['rcshow'] = self._items(show, levels=self.SHOW)
        params['rclimit'] = none_or(limit, int)
        params['rctype'] = self._items(type, self.TYPES)
        params['rctoponly'] = none_or(toponly, bool)
        params['rccontinue'] = self._check_rccontinue(rccontinue)

        doc = self.session.get(params)

        try:
            rc_docs = doc['query']['recentchanges']

            if 'query-continue' in doc:
                rccontinue = \
                        doc['query-continue']['recentchanges']['rccontinue']
            elif len(rc_docs) > 0:
                rccontinue = "|".join([rc_docs[-1]['timestamp'],
                                       str(rc_docs[-1]['rcid'] + 1)])
            else:
                pass  # Leave it be

        except KeyError as e:
            raise MalformedResponse(str(e), doc)

        return rc_docs, rccontinue



import logging

from ..util import api
from .collections import (DeletedRevs, Pages, RecentChanges, Revisions,
                          SiteInfo, UserContribs, Users)
from .errors import APIError, AuthenticationError, MalformedResponse

logger = logging.getLogger("mw.api.session")

DEFAULT_USER_AGENT = "MediaWiki-Utilities"
"""
The default User-Agent to be sent with requests to the API.
"""

class Session(api.Session):
    """
    Represents a connection to a MediaWiki API.
    
    Cookies and other session information is preserved.
    
    :Parameters:
        uri : str
            The base URI for the API to use.  Usually ends in "api.php"
        user_agent : str
            The User-Agent to be sent with requests.  Will raise a warning if
            left to default value.
    """
    
    def __init__(self, uri, *args, user_agent=DEFAULT_USER_AGENT, **kwargs):
        """
        Constructs a new :class:`Session`.
        """
        
        if user_agent == DEFAULT_USER_AGENT:
            logger.warning("Sending requests with default User-Agent.  "  +
                           "Set 'user_agent' on api.Session to quiet this " +
                           "message.")
        
        if 'headers' in kwargs:
            kwargs['headers']['User-Agent'] = str(user_agent)
        else:
            kwargs['headers'] = {'User-Agent': str(user_agent)}
        
        super().__init__(uri, *args, **kwargs)
        
        self.pages = Pages(self)
        """
        An instance of :class:`mw.api.Pages`.
        """
        
        self.revisions = Revisions(self)
        """
        An instance of :class:`mw.api.Revisions`.
        """
        
        self.recent_changes = RecentChanges(self)
        """
        An instance of :class:`mw.api.RecentChanges`.
        """
        
        self.site_info = SiteInfo(self)
        """
        An instance of :class:`mw.api.SiteInfo`.
        """
        
        self.user_contribs = UserContribs(self)
        """
        An instance of :class:`mw.api.UserContribs`.
        """
        
        self.users = Users(self)
        """
        An instance of :class:`mw.api.Users`.
        """
        
        self.deleted_revs = DeletedRevs(self)
        """
        An instance of :class:`mw.api.DeletedRevs`.
        """
        
    def login(self, username, password, token=None):
        """
        Performs a login operation.  This method usually makes two requests to
        API -- one to get a token and one to use the token to log in.  If
        authentication fails, this method will throw an
        :class:`.errors.AuthenticationError`.
        
        :Parameters:
            username : str
                Your username
            password : str
                Your password
        
        :Returns:
            The response in a json :py:class:`dict`
        """

        doc = self.post(
            {
                'action': "login",
                'lgname': username,
                'lgpassword': password,
                'lgtoken': token, # If None, we'll be getting a token
            }
        )
            
        
        try:
            if doc['login']['result'] == "Success":
                return doc
            elif doc['login']['result'] == "NeedToken":
                
                if token is not None:
                    # Woops.  We've been here before.  Better error out.
                    raise AuthenticationError(doc)
                else:
                    token = doc['login']['token']
                    return self.login(username, password, token=token)
            else:
                raise AuthenticationError(doc)
            
        except KeyError() as e:
            raise MalformedResponse(e.message, doc)
        
    
    def request(self, type, params, **kwargs):
        params.update({'format': "json"})
        
        doc = super().request(type, params, **kwargs).json()
        
        if 'error' in doc:
            raise APIError(doc)
        
        return doc



from ...types import serializable


class Comment(str, serializable.Type):
    """
    A revision comment.  This class behaves identically to
    :class:`str` except that it takes and stores an additional parameter
    recording whether the comment was deleted or not.

    >>> from mw.xml_dump import Comment
    >>>
    >>> c = Comment("foo")
    >>> c == "foo"
    True
    >>> c.deleted
    False

    **deleted**
        Was the comment deleted? | `bool`

    """

    def __new__(cls, string_or_comment="", deleted=False):
        if isinstance(string_or_comment, cls):
            return string_or_comment
        else:
            inst = super().__new__(cls, string_or_comment)
            inst.initialize(string_or_comment, deleted)
            return inst

    def initialize(self, string, deleted):
        self.deleted = bool(deleted)

    def __str__(self):
        return str.__str__(self)

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__,
            ", ".join([
                str.__repr__(self),
                "deleted={0}".format(repr(self.deleted))
            ])
        )

    def serialize(self):
        return {
            "string_or_comment": str(self),
            "deleted": self.deleted
        }

    @classmethod
    def from_element(cls, e):
        return cls(e.text, e.attr('deleted', False))



from nose.tools import eq_

from .. import ordered


def test_circle():
    circle = ordered.Circle(3)

    eq_(0, len(circle))
    print(circle.state())
    eq_(None, circle.append(5))
    eq_(1, len(circle))
    print(circle.state())
    eq_(None, circle.append(6))
    eq_(2, len(circle))
    print(circle.state())
    eq_(None, circle.append(7))
    eq_(3, len(circle))
    print(circle.state())
    eq_(5, circle.append(8))
    eq_(3, len(circle))
    print(circle.state())

    eq_([6, 7, 8], list(circle))

    print(circle.state())
    eq_([8, 7, 6], list(reversed(circle)))


def test_historical_map():
    hist = ordered.HistoricalMap(maxlen=2)

    assert "foo" not in hist

    eq_(None, hist.insert('foo', "bar1"))

    assert "foo" in hist

    eq_(None, hist.insert('foo', "bar2"))

    eq_(('foo', "bar1"), hist.insert('not_foo', "not_bar"))



from ... import types


class Namespace(types.Namespace):
    @classmethod
    def from_element(cls, element):
        return cls(
            element.attr('key'),
            element.text,
            case=element.attr('case')
        )



import logging

from ...util import none_or
from .collection import Collection

logger = logging.getLogger("mw.database.collections.pages")


class Pages(Collection):
    def get(self, page_id=None, namespace_title=None, rev_id=None):
        """
        Gets a single page based on a legitimate identifier of the page.  Note
        that namespace_title expects a tuple of namespace ID and title.

        :Parameters:
            page_id : int
                Page ID
            namespace_title : ( int, str )
                the page's namespace ID and title
            rev_id : int
                a revision ID included in the page's history

        :Returns:
            iterator over result rows
        """

        page_id = none_or(page_id, int)
        namespace_title = none_or(namespace_title, tuple)
        rev_id = none_or(rev_id, int)

        query = """
        SELECT page.*
        FROM page
        """
        values = []

        if page_id is not None:
            query += """
                WHERE page_id = %s
            """
            values.append(page_id)

        if namespace_title is not None:
            namespace, title = namespace_title

            query += " WHERE page_namespace = %s and page_title = %s "
            values.extend([int(namespace), str(title)])

        elif rev_id is not None:
            query += """
                WHERE page_id = (SELECT rev_page FROM revision WHERE rev_id = %s)
            """
            values.append(rev_id)

        else:
            raise TypeError("Must specify a page identifier.")

        cursor = self.db.shared_connection.cursor()
        cursor.execute(
            query,
            values
        )

        for row in cursor:
            return row



from itertools import chain

from nose.tools import eq_
from .. import defaults
from ..functions import sessions


EVENTS = {
    "foo": [
        [
            ("foo", 1234567890, 1),
            ("foo", 1234567892, 2),
            ("foo", 1234567894, 3)
        ],
        [
            ("foo", 1234567894 + defaults.CUTOFF, 4),
            ("foo", 1234567897 + defaults.CUTOFF, 5)
        ]
    ],
    "bar": [
        [
            ("bar", 1234567891, 6),
            ("bar", 1234567892, 7),
            ("bar", 1234567893, 8)
        ],
        [
            ("bar", 1234567895 + defaults.CUTOFF, 9),
            ("bar", 1234567898 + defaults.CUTOFF, 0)
        ]
    ]
}


def test_group_events():
    events = []
    events.extend(chain(*EVENTS['foo']))
    events.extend(chain(*EVENTS['bar']))

    events.sort()

    user_sessions = sessions(events)

    counts = {
        'foo': 0,
        'bar': 0
    }

    for user, session in user_sessions:
        eq_(list(e[2] for e in EVENTS[user][counts[user]]), list(session))
        counts[user] += 1



from nose.tools import eq_

from .. import autovivifying


def test_word_count():
    words = """
    I am a little teapot short and stout.  Here is my handle and here is my
    spout.  The red fox jumps over the lazy brown dog.  She sells sea shells
    by the sea shore.
    """.replace(".", " ").split()

    # Lame way
    lame_counts = {}
    for word in words:
        if word not in lame_counts:
            lame_counts[word] = 0

        lame_counts[word] += 1

    # Awesome way
    awesome_counts = autovivifying.Dict(  # Autovivifies entries with zero.
                                          vivifier=lambda k: 0  # Useful for counting.
    )
    for word in words:
        awesome_counts[word] += 1

    eq_(lame_counts, awesome_counts)



from nose.tools import eq_
from ..aggregate import aggregate


def test_group():
    l = [0, 1, 2, 3, 4, 5, 10, 11, 12, 13, 14]
    expected = [[0, 1, 2, 3, 4, 5], [10, 11, 12, 13, 14]]

    result = []
    for identifier, group in aggregate(l, lambda item: int(item / 10)):
        result.append(list(group))

    eq_(result, expected)



def normalize(title):
    """
    Normalizes a page title to the database format.  E.g. spaces are converted
    to underscores and the first character in the title is converted to
    upper-case.

    :Parameters:
        title : str
            A page title
    :Returns:
        The normalized title.
    :Example:
        >>> from mw.lib import title
        >>>
        >>> title.normalize("foo bar")
        'Foo_bar'

    """
    if title is None:
        return title
    else:
        if len(title) > 0:
            return (title[0].upper() + title[1:]).replace(" ", "_")
        else:
            return ""



import logging

from ...util import none_or
from .collection import Collection

logger = logging.getLogger("mw.api.collections.pages")


class Pages(Collection):
    """
    TODO
    """

    def _edit(self, title=None, pageid=None, section=None, sectiontitle=None,
              text=None, token=None, summary=None, minor=None,
              notminor=None, bot=None, basetimestamp=None,
              starttimestamp=None, recreate=None, createonly=None,
              nocreate=None, watch=None, unwatch=None, watchlist=None,
              md5=None, prependtext=None, appendtext=None, undo=None,
              undoafter=None, redirect=None, contentformat=None,
              contentmodel=None, assert_=None, nassert=None,
              captchaword=None, captchaid=None):
        params = {
            'action': "edit"
        }
        params['title'] = none_or(title, str)
        params['pageid'] = none_or(pageid, int)
        params['section'] = none_or(section, int, levels={'new'})
        params['sectiontitle'] = none_or(sectiontitle, str)
        params['text'] = none_or(text, str)
        params['token'] = none_or(token, str)
        params['summary'] = none_or(summary, str)
        params['minor'] = none_or(minor, bool)
        params['notminor'] = none_or(notminor, bool)
        params['bot'] = none_or(bot, bool)
        params['basetimestamp'] = self._check_timestamp(basetimestamp)
        params['starttimestamp'] = self._check_timestamp(starttimestamp)
        params['recreate'] = none_or(recreate, bool)
        params['createonly'] = none_or(createonly, bool)
        params['nocreate'] = none_or(nocreate, bool)
        params['watch'] = none_or(watch, bool)
        params['unwatch'] = none_or(unwatch, bool)
        params['watchlist'] = none_or(watchlist, bool)
        params['md5'] = none_or(md5, str)
        params['prependtext'] = none_or(prependtext, str)
        params['appendtext'] = none_or(appendtext, str)
        params['undo'] = none_or(undo, int)
        params['undoafter'] = none_or(undoafter, int)

        # TODO finish this



from nose.tools import eq_

from ..comment import Comment


def test_immutability():
    c = Comment("foo")
    b = Comment(c)
    eq_(id(c), id(b))


def test_empty_constructor():
    c = Comment()
    eq_(c, "")
    eq_(c.deleted, False)


def test_deleted_constructor():
    c = Comment("", deleted=True)
    eq_(c, "")
    eq_(c.deleted, True)


def test_full_constructor():
    c = Comment("Foobar!", deleted=False)
    eq_(c, "Foobar!")
    eq_(c.deleted, False)


def test_serialize():
    c = Comment("Foobar!", deleted=False)
    c2 = Comment.deserialize(c.serialize())
    eq_(c2.deleted, False)



import os.path

from nose.tools import eq_, raises

from ..functions import open_file


def test_open_file_7z():
    f = open_file(os.path.join(os.path.dirname(__file__), "test.7z"))
    eq_(f.read(), b"foobartest\n")



from itertools import chain

from . import defaults
from ...types import Timestamp
from ...util import none_or
from .functions import detect


def check_rev(session, rev, **kwargs):
    """
    Checks whether a revision (database row) was reverted (identity) and returns
    a named tuple of Revert(reverting, reverteds, reverted_to).

    :Parameters:
        session : :class:`mw.api.Session`
            An API session to make use of
        rev : dict
            a revision dict containing 'revid' and 'page.id'
        radius : int
            the maximum number of revisions that can be reverted
        before : :class:`mw.Timestamp`
            if set, limits the search for *reverting* revisions to those which were saved before this timestamp
        properties : set( str )
            a set of properties to include in revisions (see :class:`mw.api.Revisions`)
    """

    # extract rev_id, sha1, page_id
    if 'revid' in rev:
        rev_id = rev['revid']
    else:
        raise TypeError("rev must have 'rev_id'")
    if 'page' in rev:
        page_id = rev['page']['id']
    elif 'pageid' in rev:
        page_id = rev['pageid']
    else:
        raise TypeError("rev must have 'page' or 'pageid'")

    # run the regular check
    return check(session, rev_id, page_id=page_id, **kwargs)


def check(session, rev_id, page_id=None, radius=defaults.RADIUS,
          before=None, properties=None):
    """
    Checks whether a revision was reverted (identity) and returns a named tuple
    of Revert(reverting, reverteds, reverted_to).

    :Parameters:
        session : :class:`mw.api.Session`
            An API session to make use of
        rev_id : int
            the ID of the revision to check
        page_id : int
            the ID of the page the revision occupies (slower if not provided)
        radius : int
            the maximum number of revisions that can be reverted
        before : :class:`mw.Timestamp`
            if set, limits the search for *reverting* revisions to those which were saved before this timestamp
        properties : set( str )
            a set of properties to include in revisions (see :class:`mw.api.Revisions`)
    """

    if not hasattr(session, "revisions"):
        raise TypeError("session wrong type.  Expected a mw.api.Session.")

    rev_id = int(rev_id)
    radius = int(radius)
    page_id = none_or(page_id, int)
    before = none_or(before, Timestamp)
    properties = set(properties) if properties is not None else set()

    # If we don't have the page_id, we're going to need to look them up
    if page_id is None:
        rev = session.revisions.get(rev_id, properties={'ids'})
        page_id = rev['page']['pageid']

    # Load history and current rev
    current_and_past_revs = list(session.revisions.query(
        pageids={page_id},
        limit=radius + 1,
        start_id=rev_id + 1,  # Ensures that we capture the current revision
        direction="older",
        properties={'ids', 'timestamp', 'sha1'} | properties
    ))

    try:
        # Extract current rev and reorder history
        current_rev, past_revs = (
            current_and_past_revs[0],  # Current rev is the first one returned
            reversed(current_and_past_revs[1:])  # The rest are past revs, but they are in the wrong order
        )
    except IndexError:
        # Only way to get here is if there isn't enough history.  Couldn't be
        # reverted.  Just return None.
        return None

    # Load future revisions
    future_revs = session.revisions.query(
        pageids={page_id},
        limit=radius,
        start_id=rev_id,
        end=before,
        direction="newer",
        properties={'ids', 'timestamp', 'sha1'} | properties
    )

    # Convert to an iterable of (checksum, rev) pairs for detect() to consume
    checksum_revisions = chain(
        ((rev['sha1'], rev) for rev in past_revs),
        [(current_rev['sha1'], current_rev)],
        ((rev['sha1'], rev) for rev in future_revs)
    )

    for revert in detect(checksum_revisions, radius=radius):
        # Check that this is a relevant revert
        if rev_id in [rev['revid'] for rev in revert.reverteds]:
            return revert

    return None



from hashlib import sha1

from . import defaults
from .. import reverts
from .tokens import Token, Tokens


class Version:
    __slots__ = ('tokens')

    def __init__(self):
        self.tokens = None


class State:
    """
    Represents the state of word persistence in a page.
    See `<https://meta.wikimedia.org/wiki/Research:Content_persistence>`_

    :Parameters:
        tokenize : function( `str` ) --> list( `str` )
            A tokenizing function
        diff : function(list( `str` ), list( `str` )) --> list( `ops` )
            A function to perform a difference between token lists
        revert_radius : int
            the maximum revision distance that a revert can span.
        revert_detector : :class:`mw.lib.reverts.Detector`
            a revert detector to start process with
    :Example:
        >>> from pprint import pprint
        >>> from mw.lib import persistence
        >>>
        >>> state = persistence.State()
        >>>
        >>> pprint(state.process("Apples are red.", revision=1))
        ([Token(text='Apples', revisions=[1]),
          Token(text=' ', revisions=[1]),
          Token(text='are', revisions=[1]),
          Token(text=' ', revisions=[1]),
          Token(text='red', revisions=[1]),
          Token(text='.', revisions=[1])],
         [Token(text='Apples', revisions=[1]),
          Token(text=' ', revisions=[1]),
          Token(text='are', revisions=[1]),
          Token(text=' ', revisions=[1]),
          Token(text='red', revisions=[1]),
          Token(text='.', revisions=[1])],
         [])
        >>> pprint(state.process("Apples are blue.", revision=2))
        ([Token(text='Apples', revisions=[1, 2]),
          Token(text=' ', revisions=[1, 2]),
          Token(text='are', revisions=[1, 2]),
          Token(text=' ', revisions=[1, 2]),
          Token(text='blue', revisions=[2]),
          Token(text='.', revisions=[1, 2])],
         [Token(text='blue', revisions=[2])],
         [Token(text='red', revisions=[1])])
        >>> pprint(state.process("Apples are red.", revision=3)) # A revert!
        ([Token(text='Apples', revisions=[1, 2, 3]),
          Token(text=' ', revisions=[1, 2, 3]),
          Token(text='are', revisions=[1, 2, 3]),
          Token(text=' ', revisions=[1, 2, 3]),
          Token(text='red', revisions=[1, 3]),
          Token(text='.', revisions=[1, 2, 3])],
         [],
         [])
    """

    def __init__(self, tokenize=defaults.TOKENIZE, diff=defaults.DIFF,
                 revert_radius=reverts.defaults.RADIUS,
                 revert_detector=None):
        self.tokenize = tokenize
        self.diff = diff

        # Either pass a detector or the revert radius so I can make one
        if revert_detector is None:
            self.revert_detector = reverts.Detector(int(revert_radius))
        else:
            self.revert_detector = revert_detector

        # Stores the last tokens
        self.last = None

    def process(self, text, revision=None, checksum=None):
        """
        Modifies the internal state based a change to the content and returns
        the sets of words added and removed.

        :Parameters:
            text : str
                The text content of a revision
            revision : `mixed`
                Revision meta data
            checksum : str
                A checksum hash of the text content (will be generated if not provided)

        :Returns:
            Three :class:`~mw.lib.persistence.Tokens` lists

            current_tokens : :class:`~mw.lib.persistence.Tokens`
                A sequence of :class:`~mw.lib.persistence.Token` for the
                processed revision
            tokens_added : :class:`~mw.lib.persistence.Tokens`
                A set of tokens that were inserted by the processed revision
            tokens_removed : :class:`~mw.lib.persistence.Tokens`
                A sequence of :class:`~mw.lib.persistence.Token` removed by the
                processed revision

        """
        if checksum is None:
            checksum = sha1(bytes(text, 'utf8')).hexdigest()

        version = Version()

        revert = self.revert_detector.process(checksum, version)
        if revert is not None:  # Revert

            # Empty words.
            tokens_added = Tokens()
            tokens_removed = Tokens()

            # Extract reverted_to revision
            _, _, reverted_to = revert
            version.tokens = reverted_to.tokens

        else:

            if self.last is None:  # First version of the page!

                version.tokens = Tokens(Token(t) for t in self.tokenize(text))
                tokens_added = version.tokens
                tokens_removed = Tokens()

            else:

                # NOTICE: HEAVY COMPUTATION HERE!!!
                #
                # OK.  It's not that heavy.  It's just performing a diff,
                # but you're still going to spend most of your time here.
                # Diffs usually run in O(n^2) -- O(n^3) time and most tokenizers
                # produce a lot of tokens.
                version.tokens, tokens_added, tokens_removed = \
                    self.last.tokens.compare(self.tokenize(text), self.diff)

        version.tokens.persist(revision)

        self.last = version

        return version.tokens, tokens_added, tokens_removed



import logging
import traceback
from collections import namedtuple
from multiprocessing import Process
from queue import Empty

from .functions import open_file
from .iteration import Iterator

logger = logging.getLogger("mw.dump.processor")

ErrorItem = namedtuple("ErrorItem", ['error', 'item'])


class Processor(Process):
    def __init__(self, pathq, outputq, process_dump,
                       callback=lambda: None, logger=logger):
        self.pathq = pathq
        self.outputq = outputq
        self.process_dump = process_dump
        self.callback = callback
        self.logger = logger
        Process.__init__(self)

    def run(self):
        try:
            while True:

                # Force the queue to reset & behave reasonably
                foo = self.pathq.qsize()
                path = self.pathq.get(block=False)
                dump = Iterator.from_file(open_file(path))
                logger.info("Beginning to process {0}.".format(repr(path)))
                try:
                    for out in self.process_dump(dump, path):
                        self.outputq.put(ErrorItem(False, out))
                except Exception as error:

                    self.outputq.put(ErrorItem(True, (error, path)))

                    logger.error(
                        "Failed while processing dump " +
                        "{0}: {1}".format(repr(path),
                                          "\n" + traceback.format_exc()))
        except Empty:
            self.logger.info("Nothing left to do.  Shutting down thread.")
        finally:
            self.callback()



import logging
import sys

from ...util import none_or
from ..errors import MalformedResponse
from .collection import Collection


logger = logging.getLogger("mw.api.collections.deletedrevs")


class DeletedRevs(Collection):
    PROPERTIES = {'revid', 'parentid', 'user', 'userid', 'comment',
                  'parsedcomment', 'minor', 'len', 'sha1', 'content', 'token',
                  'tags'}

    # TODO:
    # This is *not* the right way to do this, but it should work for all queries.
    MAX_REVISIONS = 500

    def query(self, *args, limit=sys.maxsize, **kwargs):
        """
        Queries deleted revisions.
        See https://www.mediawiki.org/wiki/API:Deletedrevs

        :Parameters:
            titles : set(str)
                A set of page names to query (note that namespace prefix is expected)
            start : :class:`mw.Timestamp`
                A timestamp to start querying from
            end : :class:`mw.Timestamp`
                A timestamp to end querying
            from_title : str
                A title from which to start querying (alphabetically)
            to_title : str
                A title from which to stop querying (alphabetically)
            prefix : str
                A title prefix to match on
            drcontinue : str
                When more results are available, use this to continue (3) Note: may only work if drdir is set to newer.
            unique : bool
                List only one revision for each page
            tag : str
                Only list revision tagged with this tag
            user : str
                Only list revisions saved by this user_text
            excludeuser : str
                Do not list revision saved by this user_text
            namespace : int
                Only list pages in this namespace (id)
            limit : int
                Limit the number of results
            direction : str
                "newer" or "older"
            properties : set(str)
                A list of properties to include in the results:

                * revid          - Adds the revision ID of the deleted revision
                * parentid       - Adds the revision ID of the previous revision to the page
                * user           - Adds the user who made the revision
                * userid         - Adds the user ID whom made the revision
                * comment        - Adds the comment of the revision
                * parsedcomment  - Adds the parsed comment of the revision
                * minor          - Tags if the revision is minor
                * len            - Adds the length (bytes) of the revision
                * sha1           - Adds the SHA-1 (base 16) of the revision
                * content        - Adds the content of the revision
                * token          - Gives the edit token
                * tags           - Tags for the revision
        """
        # `limit` means something diffent here
        kwargs['limit'] = min(limit, self.MAX_REVISIONS)
        revisions_yielded = 0
        done = False
        while not done and revisions_yielded <= limit:
            rev_docs, drcontinue = self._query(*args, **kwargs)
            for doc in rev_docs:
                yield doc
                revisions_yielded += 1
                if revisions_yielded >= limit:
                    break

            if drcontinue is not None and len(rev_docs) > 0:
                kwargs['drcontinue'] = drcontinue
            else:
                done = True

    def _query(self, titles=None,
               start=None, end=None, from_title=None, to_title=None,
               prefix=None, drcontinue=None, unique=None, tag=None,
               user=None, excludeuser=None, namespace=None, limit=None,
               properties=None, direction=None):

        params = {
            'action': "query",
            'list': "deletedrevs"
        }

        params['titles'] = self._items(titles)
        params['drprefix'] = none_or(prefix, str)
        params['drfrom'] = none_or(from_title, str)
        params['drto'] = none_or(to_title, str)

        params['drprop'] = self._items(properties, levels=self.PROPERTIES)
        params['drlimit'] = none_or(limit, int)
        params['drstart'] = str(Timestamp(start))
        params['drend'] = str(Timestamp(end))

        params['drdir'] = self._check_direction(direction)
        params['druser'] = none_or(user, str)
        params['drexcludeuser'] = none_or(excludeuser, int)
        params['drtag'] = none_or(tag, str)
        params['drcontinue'] = none_or(drcontinue, str)

        doc = self.session.get(params)

        try:
            if 'query-continue' in doc:
                drcontinue = doc['query-continue']['deletedrevs']['drcontinue']
            else:
                drcontinue = None

            pages = doc['query']['deletedrevs']
            rev_docs = []

            for page_doc in pages:
                page_rev_docs = page_doc['revisions']
                del page_doc['revisions']

                for rev_doc in page_rev_docs:
                    rev_doc['page'] = page_doc

                rev_docs.extend(page_rev_docs)

            return rev_docs, drcontinue

        except KeyError as e:
            print(doc)
            raise MalformedResponse(str(e), doc)



from nose.tools import eq_

from ..cache import Cache


def test_session_manager():
    cache = Cache(cutoff=2)

    user_sessions = list(cache.process("foo", 1))
    eq_(user_sessions, [])

    user_sessions = list(cache.process("bar", 2))
    eq_(user_sessions, [])

    user_sessions = list(cache.process("foo", 2))
    eq_(user_sessions, [])

    user_sessions = list(cache.process("bar", 10))
    eq_(len(user_sessions), 2)

    user_sessions = list(cache.get_active_sessions())
    eq_(len(user_sessions), 1)



from ...types import serializable
from ...util import none_or

from .util import consume_tags


class Contributor(serializable.Type):
    """
    Contributor meta data.
    """
    __slots__ = ('id', 'user_text')

    TAG_MAP = {
        'id': lambda e: int(e.text),
        'username': lambda e: str(e.text),
        'ip': lambda e: str(e.text)
    }

    def __init__(self, id, user_text):
        self.id = none_or(id, int)
        """
        User ID : int | `None` (if not specified in the XML)

        User ID of a user if the contributor is signed into an account
        in the while making the contribution and `None` when
        contributors are not signed in.
        """

        self.user_text = none_or(user_text, str)
        """
        User name or IP address : str | `None` (if not specified in the XML)

        If a user is logged in, this will reflect the users accout
        name. If the user is not logged in, this will usually be
        recorded as the IPv4 or IPv6 address in the XML.
        """

    @classmethod
    def from_element(cls, element):
        values = consume_tags(cls.TAG_MAP, element)

        return cls(
            values.get('id'),
            values.get('username', values.get('ip'))
        )



from ...types import Namespace
from ...util import autovivifying, none_or

from .functions import normalize


class Parser:
    """
    Constructs a page name parser from a set of :class:`mw.Namespace`.  Such a
    parser can be used to convert a full page name (namespace included with a
    colon; e.g, ``"Talk:Foo"``) into a namespace ID and
    :func:`mw.lib.title.normalize`'d page title (e.g., ``(1, "Foo")``).

    :Parameters:
        namespaces : set( :class:`mw.Namespace` )
    :Example:
        >>> from mw import Namespace
        >>> from mw.lib import title
        >>>
        >>> parser = title.Parser(
        ...     [
        ...             Namespace(0, "", case="first-letter"),
        ...             Namespace(1, "Discuss\u00e3o", canonical="Talk", case="first-letter"),
        ...             Namespace(2, "Usu\u00e1rio(a)", canonical="User", aliases={"U"}, case="first-letter")
        ...     ]
        ... )
        >>>
        >>> parser.parse("Discuss\u00e3o:Foo") # Using the standard name
        (1, 'Foo')
        >>> parser.parse("Talk:Foo bar") # Using the cannonical name
        (1, 'Foo_bar')
        >>> parser.parse("U:Foo bar") # Using an alias
        (2, 'Foo_bar')
        >>> parser.parse("Herpderp:Foo bar") # Psuedo namespace
        (0, 'Herpderp:Foo_bar')
    """

    def __init__(self, namespaces=None):
        namespaces = none_or(namespaces, set)

        self.ids = {}
        self.names = {}

        if namespaces is not None:
            for namespace in namespaces:
                self.add_namespace(namespace)

    def parse(self, page_name):
        """
        Parses a page name to extract the namespace.

        :Parameters:
            page_name : str
                A page name including the namespace prefix and a colon (if not Main)

        :Returns:
            A tuple of (namespace : `int`, title : `str`)
        """
        parts = page_name.split(":", 1)
        if len(parts) == 1:
            ns_id = 0
            title = normalize(page_name)
        else:
            ns_name, title = parts
            ns_name, title = normalize(ns_name), normalize(title)

            if self.contains_name(ns_name):
                ns_id = self.get_namespace(name=ns_name).id
            else:
                ns_id = 0
                title = normalize(page_name)

        return ns_id, title

    def add_namespace(self, namespace):
        """
        Adds a namespace to the parser.

        :Parameters:
            namespace : :class:`mw.Namespace`
                A namespace
        """
        self.ids[namespace.id] = namespace
        self.names[namespace.name] = namespace

        for alias in namespace.aliases:
            self.names[alias] = namespace

        if namespace.canonical is not None:
            self.names[namespace.canonical] = namespace

    def contains_name(self, name):
        return normalize(name) in self.names

    def get_namespace(self, id=None, name=None):
        """
        Gets a namespace from the parser.  Throws a :class:`KeyError` if a
        namespace cannot be found.

        :Parameters:
            id : int
                A namespace ID
            name : str
                A namespace name (standard, cannonical names and aliases
                will be searched)
        :Returns:
            A :class:`mw.Namespace`.
        """
        if id is not None:
            return self.ids[int(id)]
        else:
            return self.names[normalize(name)]

    @classmethod
    def from_site_info(cls, si_doc):
        """
        Constructs a parser from the result of a :meth:`mw.api.SiteInfo.query`.

        :Parameters:
            si_doc : dict
                The result of a site_info request.

        :Returns:
            An initialized :class:`mw.lib.title.Parser`
        """
        aliases = autovivifying.Dict(vivifier=lambda k: [])
        # get aliases
        if 'namespacealiases' in si_doc:
            for alias_doc in si_doc['namespacealiases']:
                aliases[alias_doc['id']].append(alias_doc['*'])

        namespaces = []
        for ns_doc in si_doc['namespaces'].values():
            namespaces.append(
                Namespace(
                    ns_doc['id'],
                    ns_doc['*'],
                    canonical=ns_doc.get('canonical'),
                    aliases=aliases[ns_doc['id']],
                    case=ns_doc['case']
                )
            )

        return Parser(namespaces)

    @classmethod
    def from_api(cls, session):
        """
        Constructs a parser from a :class:`mw.api.Session`

        :Parameters:
            session : :class:`mw.api.Session`
                An open API session

        :Returns:
            An initialized :class:`mw.lib.title.Parser`
        """
        si_doc = session.site_info.query(
            properties={'namespaces', 'namespacealiases'}
        )

        return cls.from_site_info(si_doc)

    @classmethod
    def from_dump(cls, dump):
        """
        Constructs a parser from a :class:`mw.xml_dump.Iterator`.  Note that
        XML database dumps do not include namespace aliases or cannonical names
        so the parser that will be constructed will only work in common cases.

        :Parameters:
            dump : :class:`mw.xml_dump.Iterator`
                An XML dump iterator

        :Returns:
            An initialized :class:`mw.lib.title.Parser`
        """
        return cls(dump.namespaces)



from nose.tools import eq_

from ....types import Namespace

from ..parser import Parser


def test_simple():
    parser = Parser(
        [
            Namespace(0, "", case="first-letter"),
            Namespace(1, "Discuss\u00e3o", canonical="Talk", case="first-letter"),
            Namespace(2, "Usu\u00e1rio(a)", canonical="User", case="first-letter")
        ]
    )

    eq_((1, "Foo"), parser.parse("Discuss\u00e3o:Foo"))
    eq_((1, "Foo_bar"), parser.parse("Discuss\u00e3o:Foo bar"))
    eq_((0, "Herpderp:Foo_bar"), parser.parse("Herpderp:Foo bar"))


def test_from_site_info():
    parser = Parser.from_site_info(
        {
            "namespaces": {
                "0": {
                    "id": 0,
                    "case": "first-letter",
                    "*": "",
                    "content": ""
                },
                "1": {
                    "id": 1,
                    "case": "first-letter",
                    "*": "Discuss\u00e3o",
                    "subpages": "",
                    "canonical": "Talk"
                },
                "2": {
                    "id": 2,
                    "case": "first-letter",
                    "*": "Usu\u00e1rio(a)",
                    "subpages": "",
                    "canonical": "User"
                }
            }
        }
    )

    eq_((1, "Foo"), parser.parse("Discuss\u00e3o:Foo"))
    eq_((1, "Foo_bar"), parser.parse("Discuss\u00e3o:Foo bar"))
    eq_((0, "Herpderp:Foo_bar"), parser.parse("Herpderp:Foo bar"))



import logging
from collections import namedtuple

logger = logging.getLogger("mw.lib.sessions.event")


# class Event:
#   __slots__ = ('user', 'timestamp', 'data')
#
#   def __init__(self, user, timestamp, data=None):
#       self.user = user
#       self.timestamp = Timestamp(timestamp)
#       self.data = data

Event = namedtuple("Event", ['user', 'timestamp', 'data'])


def unpack_events(events):
    return list(e.data for e in events)



"""
This is a failed attempt.  See
https://github.com/halfak/Mediawiki-Utilities/issues/13 for more details.
"""

'''
import os

import py7zlib


class SevenZFileError(py7zlib.ArchiveError):
    pass

class SevenZFile(object):
    @classmethod
    def is_7zfile(cls, filepath):
        """ Determine if filepath points to a valid 7z archive. """
        is7z = False
        fp = None
        try:
            fp = open(filepath, 'rb')
            archive = py7zlib.Archive7z(fp)
            n = len(archive.getnames())
            is7z = True
        finally:
            if fp: fp.close()
        return is7z

    def __init__(self, filepath):
        fp = open(filepath, 'rb')
        self.filepath = filepath
        self.archive = py7zlib.Archive7z(fp)

    def __contains__(self, name):
        return name in self.archive.getnames()

    def bytestream(self, name):
        """ Iterate stream of bytes from an archive member. """
        if name not in self:
            raise SevenZFileError('member %s not found in %s' %
                                  (name, self.filepath))
        else:
            member = self.archive.getmember(name)
            for byte in member.read():
                if not byte: break
                yield byte

    def readlines(self, name):
        """ Iterate lines from an archive member. """
        linesep = os.linesep[-1]
        line = ''
        for ch in self.bytestream(name):
            line += ch
            if ch == linesep:
                yield line
                line = ''
        if line: yield line
        
    
import os

import py7zlib

with open("/mnt/data/xmldatadumps/public/simplewiki/20141122/simplewiki-20141122-pages-meta-history.xml.7z", "rb") as f:
    a = py7zlib.Archive7z(f)
    
    print(a.getmember(a.getnames()[0]).read())
'''
